from __future__ import annotations

import re
from typing import Dict

BASE64_ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
BASE64_LOOKUP: Dict[str, int] = {ch: idx for idx, ch in enumerate(BASE64_ALPHABET)}


def utf8_encode(input_str: str) -> bytes:
    return input_str.encode("utf-8")


def utf8_decode(input_bytes: bytes) -> str:
    return input_bytes.decode("utf-8")


def encode_base64(data: bytes) -> str:
    out = []
    for i in range(0, len(data), 3):
        b0 = data[i] if i < len(data) else 0
        b1 = data[i + 1] if i + 1 < len(data) else 0
        b2 = data[i + 2] if i + 2 < len(data) else 0
        n = (b0 << 16) | (b1 << 8) | b2

        out.append(BASE64_ALPHABET[(n >> 18) & 63])
        out.append(BASE64_ALPHABET[(n >> 12) & 63])
        out.append(BASE64_ALPHABET[(n >> 6) & 63] if i + 1 < len(data) else "=")
        out.append(BASE64_ALPHABET[n & 63] if i + 2 < len(data) else "=")

    return "".join(out)


def decode_base64(base64: str) -> bytes:
    clean = re.sub(r"\s+", "", base64)
    if len(clean) % 4 != 0:
        raise ValueError("Invalid base64 length")

    out = bytearray()

    for i in range(0, len(clean), 4):
        c0, c1, c2, c3 = clean[i : i + 4]

        v0 = BASE64_LOOKUP.get(c0)
        v1 = BASE64_LOOKUP.get(c1)
        v2 = 0 if c2 == "=" else BASE64_LOOKUP.get(c2)
        v3 = 0 if c3 == "=" else BASE64_LOOKUP.get(c3)

        if v0 is None or v1 is None or (c2 != "=" and v2 is None) or (c3 != "=" and v3 is None):
            raise ValueError("Invalid base64 input")

        n = (v0 << 18) | (v1 << 12) | (v2 << 6) | v3
        out.append((n >> 16) & 255)
        if c2 != "=":
            out.append((n >> 8) & 255)
        if c3 != "=":
            out.append(n & 255)

    return bytes(out)


def encode_base64url(data: bytes) -> str:
    return encode_base64(data).replace("+", "-").replace("/", "_").rstrip("=")


def decode_base64url(input_str: str) -> bytes:
    base64 = input_str.replace("-", "+").replace("_", "/")
    pad = (4 - (len(base64) % 4)) % 4
    return decode_base64(base64 + ("=" * pad))


utf8Encode = utf8_encode
utf8Decode = utf8_decode
encodeBase64 = encode_base64
decodeBase64 = decode_base64
encodeBase64Url = encode_base64url
decodeBase64Url = decode_base64url

__all__ = [
    "decode_base64",
    "decode_base64url",
    "decodeBase64",
    "decodeBase64Url",
    "encode_base64",
    "encode_base64url",
    "encodeBase64",
    "encodeBase64Url",
    "utf8_decode",
    "utf8_encode",
    "utf8Decode",
    "utf8Encode",
]



from .mastercard_vi import (
    VerifiableIntentRecord,
    aar_to_verifiable_intent,
    aarToVerifiableIntent,
    verifiable_intent_to_aar,
    verifiableIntentToAAR,
)

__all__ = [
    "VerifiableIntentRecord",
    "aar_to_verifiable_intent",
    "aarToVerifiableIntent",
    "verifiable_intent_to_aar",
    "verifiableIntentToAAR",
]



from __future__ import annotations

import pytest

from botindex_aar.encoding import (
    decode_base64,
    decode_base64url,
    encode_base64,
    encode_base64url,
    utf8_decode,
    utf8_encode,
)


def test_utf8_roundtrip() -> None:
    source = "AAR ✓ 日本語"
    assert utf8_decode(utf8_encode(source)) == source


def test_encode_decode_base64_roundtrip() -> None:
    data = b"\x00\x01\xfe\xff"
    encoded = encode_base64(data)
    assert encoded == "AAH+/w=="
    assert decode_base64(encoded) == data


def test_decode_base64_strips_whitespace() -> None:
    assert decode_base64(" YW Jj\n") == b"abc"


def test_decode_base64_invalid_length() -> None:
    with pytest.raises(ValueError, match="Invalid base64 length"):
        decode_base64("abc")


def test_decode_base64_invalid_input() -> None:
    with pytest.raises(ValueError, match="Invalid base64 input"):
        decode_base64("ab!#")


def test_encode_decode_base64url_roundtrip() -> None:
    data = b"\xfb\xff\x00"
    encoded = encode_base64url(data)
    assert encoded == "-_8A"
    assert decode_base64url(encoded) == data



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Union

from ..types import to_mapping


@dataclass
class VerifiableIntentRecord:
    recordId: str
    identity: Dict[str, Any]
    intent: Dict[str, Any]
    action: Dict[str, Any]
    outcome: Dict[str, Any]
    proof: Dict[str, Any]
    disclosure: Dict[str, Any]
    metadata: Dict[str, Any]


def aar_to_verifiable_intent(receipt: Union[Mapping[str, Any], Any]) -> Dict[str, Any]:
    r = to_mapping(receipt, "receipt")

    return {
        "recordId": r["receiptId"],
        "identity": {
            "id": r["principal"]["id"],
            "type": r["principal"]["type"],
        },
        "intent": {
            "description": "%s on %s" % (r["action"]["type"], r["action"]["target"]),
            "permissions": list(r["scope"]["permissions"]),
            "constraints": r["scope"].get("constraints"),
            "timestamp": r["timestamp"],
        },
        "action": {
            "agentId": r["agent"]["id"],
            "agentName": r["agent"].get("name"),
            "type": r["action"]["type"],
            "target": r["action"]["target"],
            "method": r["action"].get("method"),
            "status": r["action"]["status"],
            "timestamp": r["timestamp"],
        },
        "outcome": {
            "inputDigest": "%s:%s" % (r["inputHash"]["alg"], r["inputHash"]["digest"]),
            "outputDigest": "%s:%s" % (r["outputHash"]["alg"], r["outputHash"]["digest"]),
            "cost": {
                "amount": r["cost"]["amount"],
                "currency": r["cost"]["currency"],
            },
        },
        "proof": {
            "algorithm": r["signature"]["alg"],
            "keyId": r["signature"]["kid"],
            "publicKey": r["signature"].get("publicKey"),
            "signature": r["signature"]["sig"],
            "canonicalization": r["signature"]["canonicalization"],
        },
        "disclosure": {
            "mode": "minimal",
            "redactedFields": ["inputHash.raw", "outputHash.raw"],
        },
        "metadata": {**dict(r.get("metadata") or {}), "aarVersion": "1.0"},
    }


def verifiable_intent_to_aar(vi: Union[VerifiableIntentRecord, Mapping[str, Any]]) -> Dict[str, Any]:
    obj = vi.__dict__ if isinstance(vi, VerifiableIntentRecord) else dict(vi)

    outcome = obj.get("outcome") or {}
    input_digest_parts = str(outcome.get("inputDigest") or "").split(":", 1)
    output_digest_parts = str(outcome.get("outputDigest") or "").split(":", 1)

    input_alg = input_digest_parts[0] if len(input_digest_parts) > 0 and input_digest_parts[0] else "sha256"
    input_digest = input_digest_parts[1] if len(input_digest_parts) > 1 else ""

    output_alg = output_digest_parts[0] if len(output_digest_parts) > 0 and output_digest_parts[0] else "sha256"
    output_digest = output_digest_parts[1] if len(output_digest_parts) > 1 else ""

    action = obj.get("action") or {}
    identity = obj.get("identity") or {}
    intent = obj.get("intent") or {}
    proof = obj.get("proof") or {}

    raw_cost = outcome.get("cost")
    cost = (
        {
            "amount": raw_cost.get("amount"),
            "currency": raw_cost.get("currency"),
        }
        if isinstance(raw_cost, Mapping)
        else {"amount": "0", "currency": "USD"}
    )

    return {
        "receiptId": obj.get("recordId"),
        "agent": {
            "id": action.get("agentId"),
            "name": action.get("agentName"),
        },
        "principal": {
            "id": identity.get("id"),
            "type": identity.get("type"),
        },
        "action": {
            "type": action.get("type"),
            "target": action.get("target"),
            "method": action.get("method"),
            "status": action.get("status"),
        },
        "scope": {
            "permissions": list(intent.get("permissions") or []),
            "constraints": intent.get("constraints"),
        },
        "inputHash": {
            "alg": input_alg,
            "digest": input_digest,
        },
        "outputHash": {
            "alg": output_alg,
            "digest": output_digest,
        },
        "timestamp": action.get("timestamp"),
        "cost": cost,
        "signature": {
            "alg": proof.get("algorithm"),
            "kid": proof.get("keyId"),
            "publicKey": proof.get("publicKey"),
            "canonicalization": proof.get("canonicalization"),
            "sig": proof.get("signature"),
        },
        "metadata": {**dict(obj.get("metadata") or {}), "sourceFormat": "verifiable-intent"},
    }


aarToVerifiableIntent = aar_to_verifiable_intent
verifiableIntentToAAR = verifiable_intent_to_aar

__all__ = [
    "VerifiableIntentRecord",
    "aar_to_verifiable_intent",
    "aarToVerifiableIntent",
    "verifiable_intent_to_aar",
    "verifiableIntentToAAR",
]



from __future__ import annotations

from typing import Any, Dict, Mapping, Optional, Union

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from .canonicalize import canonicalize_for_signing
from .encoding import decode_base64url, utf8_encode
from .types import VerifyResult, to_mapping


def _shape_error(receipt: Mapping[str, Any]) -> Optional[str]:
    if not receipt.get("receiptId"):
        return "missing required field: receiptId"

    agent = receipt.get("agent")
    if not isinstance(agent, Mapping) or not agent.get("id"):
        return "missing required field: agent.id"

    principal = receipt.get("principal")
    if not isinstance(principal, Mapping) or not principal.get("id") or not principal.get("type"):
        return "missing required field: principal"

    action = receipt.get("action")
    if not isinstance(action, Mapping) or not action.get("type") or not action.get("target") or not action.get("status"):
        return "missing required field: action"

    scope = receipt.get("scope")
    permissions = scope.get("permissions") if isinstance(scope, Mapping) else None
    if not isinstance(permissions, list) or len(permissions) == 0:
        return "missing required field: scope.permissions"

    input_hash = receipt.get("inputHash")
    output_hash = receipt.get("outputHash")
    if not isinstance(input_hash, Mapping) or not input_hash.get("digest") or not isinstance(output_hash, Mapping) or not output_hash.get("digest"):
        return "missing required field: hash objects"

    signature = receipt.get("signature")
    if not isinstance(signature, Mapping) or not signature.get("sig"):
        return "missing signature.sig"

    if signature.get("alg") != "Ed25519":
        return "unsupported signature.alg"

    if signature.get("canonicalization") != "JCS-SORTED-UTF8-NOWS":
        return "unsupported canonicalization"

    return None


def _resolve_public_key(receipt: Mapping[str, Any], explicit: Optional[Union[bytes, bytearray, str]] = None) -> Optional[bytes]:
    if isinstance(explicit, (bytes, bytearray)):
        return bytes(explicit)

    if isinstance(explicit, str) and len(explicit) > 0:
        return decode_base64url(explicit)

    signature = receipt.get("signature")
    if isinstance(signature, Mapping) and signature.get("publicKey"):
        return decode_base64url(str(signature["publicKey"]))

    agent = receipt.get("agent")
    if isinstance(agent, Mapping) and agent.get("publicKey"):
        return decode_base64url(str(agent["publicKey"]))

    return None


def verify_receipt(receipt: Union[Mapping[str, Any], Any], public_key: Optional[Union[bytes, bytearray, str]] = None) -> VerifyResult:
    receipt_obj = to_mapping(receipt, "receipt")

    err = _shape_error(receipt_obj)
    if err:
        return VerifyResult(ok=False, reason=err)

    try:
        key = _resolve_public_key(receipt_obj, public_key)
        if not key:
            return VerifyResult(ok=False, reason="no public key provided")

        if len(key) != 32:
            return VerifyResult(ok=False, reason="invalid public key length")

        signature = receipt_obj["signature"]
        sig = decode_base64url(str(signature["sig"]))
        if len(sig) != 64:
            return VerifyResult(ok=False, reason="invalid signature length")

        msg = utf8_encode(canonicalize_for_signing(receipt_obj))

        try:
            VerifyKey(key).verify(msg, sig)
            return VerifyResult(ok=True)
        except BadSignatureError:
            return VerifyResult(ok=False, reason="signature verification failed")
    except Exception as exc:  # pragma: no cover - defensive parity with TS
        return VerifyResult(ok=False, reason=str(exc) if str(exc) else "verification error")


verifyReceipt = verify_receipt

__all__ = ["verify_receipt", "verifyReceipt"]



from .fastapi import AARMiddleware, AARMiddlewareOptions
from .flask import aar_middleware, install_aar_middleware

__all__ = [
    "AARMiddleware",
    "AARMiddlewareOptions",
    "aar_middleware",
    "install_aar_middleware",
]



from __future__ import annotations

import json
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, Mapping, Union

from .encoding import encode_base64, utf8_encode
from .sign import sign_receipt
from .types import to_mapping, to_plain


def _create_receipt_id() -> str:
    return str(uuid.uuid4())


def _iso_now() -> str:
    return datetime.now(timezone.utc).isoformat(timespec="milliseconds").replace("+00:00", "Z")


def create_receipt(opts: Union[Mapping[str, Any], Any]) -> Dict[str, Any]:
    opts_obj = to_mapping(opts, "opts")

    agent = opts_obj.get("agent")
    if not isinstance(agent, Mapping) or not agent.get("id"):
        raise ValueError("opts.agent.id is required")

    signature_opts = opts_obj.get("signature")
    if signature_opts is None:
        signature_opts = {}
    if not isinstance(signature_opts, Mapping):
        raise ValueError("opts.signature must be an object")

    signature = {
        "alg": "Ed25519",
        "canonicalization": "JCS-SORTED-UTF8-NOWS",
        "kid": signature_opts.get("kid") or (str(agent["id"]) + "#key-1"),
    }

    if "publicKey" in signature_opts and signature_opts.get("publicKey") is not None:
        signature["publicKey"] = signature_opts.get("publicKey")

    if "sig" in signature_opts and signature_opts.get("sig") is not None:
        signature["sig"] = signature_opts.get("sig")

    return {
        "receiptId": opts_obj.get("receiptId") or _create_receipt_id(),
        "agent": to_plain(opts_obj["agent"]),
        "principal": to_plain(opts_obj["principal"]),
        "action": to_plain(opts_obj["action"]),
        "scope": to_plain(opts_obj["scope"]),
        "inputHash": to_plain(opts_obj["inputHash"]),
        "outputHash": to_plain(opts_obj["outputHash"]),
        "timestamp": opts_obj.get("timestamp") or _iso_now(),
        "cost": to_plain(opts_obj["cost"]),
        "signature": signature,
        "metadata": to_plain(opts_obj.get("metadata") or {}),
    }


def sign_and_finalize(unsigned: Union[Mapping[str, Any], Any], secret_key: Union[bytes, bytearray, str]) -> Dict[str, Any]:
    return sign_receipt(unsigned, secret_key)


def encode_receipt_header(receipt: Union[Mapping[str, Any], Any]) -> str:
    receipt_obj = to_plain(receipt)
    text = json.dumps(receipt_obj, ensure_ascii=False, separators=(",", ":"))
    return encode_base64(utf8_encode(text))


createReceipt = create_receipt
signAndFinalize = sign_and_finalize
encodeReceiptHeader = encode_receipt_header

__all__ = [
    "create_receipt",
    "createReceipt",
    "sign_and_finalize",
    "signAndFinalize",
    "encode_receipt_header",
    "encodeReceiptHeader",
]



from .canonicalize import canonicalize, canonicalize_for_signing, canonicalizeForSigning
from .compat import (
    VerifiableIntentRecord,
    aar_to_verifiable_intent,
    aarToVerifiableIntent,
    verifiable_intent_to_aar,
    verifiableIntentToAAR,
)
from .discovery import DiscoveryOptions, build_well_known_config, buildWellKnownConfig, well_known_handler, wellKnownHandler
from .encoding import (
    decode_base64,
    decode_base64url,
    decodeBase64,
    decodeBase64Url,
    encode_base64,
    encode_base64url,
    encodeBase64,
    encodeBase64Url,
    utf8_decode,
    utf8_encode,
    utf8Decode,
    utf8Encode,
)
from .hash import hash_input, hash_output, hashInput, hashOutput, sha256
from .receipt import (
    create_receipt,
    createReceipt,
    encode_receipt_header,
    encodeReceiptHeader,
    sign_and_finalize,
    signAndFinalize,
)
from .sign import (
    generate_key_pair,
    generateKeyPair,
    load_secret_key,
    loadSecretKey,
    public_key_from_secret,
    publicKeyFromSecret,
    sign_receipt,
    signReceipt,
)
from .types import (
    AARReceipt,
    Action,
    ActionStatus,
    AgentIdentity,
    Cost,
    CreateReceiptOptions,
    HashAlgorithm,
    HashObject,
    JSONValue,
    KeyPair,
    Principal,
    PrincipalType,
    Scope,
    Signature,
    UnsignedReceipt,
    UnsignedSignature,
    VerifyResult,
    WellKnownAARAgent,
    WellKnownAARConfiguration,
)
from .verify import verify_receipt, verifyReceipt

__all__ = [
    "AARReceipt",
    "Action",
    "ActionStatus",
    "AgentIdentity",
    "Cost",
    "CreateReceiptOptions",
    "DiscoveryOptions",
    "HashAlgorithm",
    "HashObject",
    "JSONValue",
    "KeyPair",
    "Principal",
    "PrincipalType",
    "Scope",
    "Signature",
    "UnsignedReceipt",
    "UnsignedSignature",
    "VerifyResult",
    "VerifiableIntentRecord",
    "WellKnownAARAgent",
    "WellKnownAARConfiguration",
    "aar_to_verifiable_intent",
    "aarToVerifiableIntent",
    "build_well_known_config",
    "buildWellKnownConfig",
    "canonicalize",
    "canonicalize_for_signing",
    "canonicalizeForSigning",
    "create_receipt",
    "createReceipt",
    "decode_base64",
    "decode_base64url",
    "decodeBase64",
    "decodeBase64Url",
    "encode_base64",
    "encode_base64url",
    "encode_receipt_header",
    "encodeBase64",
    "encodeBase64Url",
    "encodeReceiptHeader",
    "generate_key_pair",
    "generateKeyPair",
    "hash_input",
    "hash_output",
    "hashInput",
    "hashOutput",
    "load_secret_key",
    "loadSecretKey",
    "public_key_from_secret",
    "publicKeyFromSecret",
    "sha256",
    "sign_and_finalize",
    "sign_receipt",
    "signAndFinalize",
    "signReceipt",
    "utf8_decode",
    "utf8_encode",
    "utf8Decode",
    "utf8Encode",
    "verifiable_intent_to_aar",
    "verifiableIntentToAAR",
    "verify_receipt",
    "verifyReceipt",
    "well_known_handler",
    "wellKnownHandler",
]



from __future__ import annotations

import pytest

from botindex_aar.canonicalize import canonicalize, canonicalize_for_signing


def test_canonicalize_sorts_keys_deep() -> None:
    value = {
        "z": 1,
        "a": {
            "d": 4,
            "c": [
                {"b": 2, "a": 1},
                3,
            ],
        },
    }
    assert canonicalize(value) == '{"a":{"c":[{"a":1,"b":2},3],"d":4},"z":1}'


def test_canonicalize_for_signing_removes_signature_sig() -> None:
    receipt = {
        "receiptId": "r1",
        "signature": {
            "alg": "Ed25519",
            "kid": "k1",
            "canonicalization": "JCS-SORTED-UTF8-NOWS",
            "sig": "abc",
        },
    }
    assert canonicalize_for_signing(receipt) == (
        '{"receiptId":"r1","signature":{"alg":"Ed25519","canonicalization":"JCS-SORTED-UTF8-NOWS","kid":"k1"}}'
    )


def test_canonicalize_for_signing_invalid_signature() -> None:
    with pytest.raises(ValueError, match="Invalid receipt.signature"):
        canonicalize_for_signing({"signature": "bad"})



from __future__ import annotations

import hashlib
from typing import Any, Union

from .canonicalize import canonicalize
from .encoding import encode_base64url, utf8_encode
from .types import HashObject


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def hash_input(data: Any) -> HashObject:
    bytes_data = utf8_encode(canonicalize(data))
    return HashObject(alg="sha256", digest=encode_base64url(sha256(bytes_data)))


def hash_output(data: Union[str, bytes, bytearray]) -> HashObject:
    bytes_data = utf8_encode(data) if isinstance(data, str) else bytes(data)
    return HashObject(alg="sha256", digest=encode_base64url(sha256(bytes_data)))


hashInput = hash_input
hashOutput = hash_output

__all__ = ["sha256", "hash_input", "hash_output", "hashInput", "hashOutput"]



from __future__ import annotations

from dataclasses import dataclass, fields, is_dataclass, field
from typing import Any, Dict, List, Literal, Mapping, Optional, Sequence, Tuple, Union

HashAlgorithm = Literal["sha256", "sha384", "sha512", "blake3"]
PrincipalType = Literal["user", "organization", "service", "agent", "other"]
ActionStatus = Literal["success", "failure", "partial"]


@dataclass
class HashObject:
    alg: HashAlgorithm
    digest: str


@dataclass
class AgentIdentity:
    id: str
    name: Optional[str] = None
    version: Optional[str] = None
    publicKey: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Principal:
    id: str
    type: PrincipalType
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Action:
    type: str
    target: str
    status: ActionStatus
    method: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Scope:
    permissions: List[str]
    constraints: Optional[Dict[str, Any]] = None
    x402: Optional[Dict[str, Any]] = None
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Cost:
    amount: str
    currency: str
    unit: Optional[str] = None
    payer: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Signature:
    kid: str
    sig: str
    alg: Literal["Ed25519"] = "Ed25519"
    publicKey: Optional[str] = None
    canonicalization: Literal["JCS-SORTED-UTF8-NOWS"] = "JCS-SORTED-UTF8-NOWS"


@dataclass
class UnsignedSignature:
    kid: str
    alg: Literal["Ed25519"] = "Ed25519"
    publicKey: Optional[str] = None
    canonicalization: Literal["JCS-SORTED-UTF8-NOWS"] = "JCS-SORTED-UTF8-NOWS"
    sig: Optional[str] = None


@dataclass
class AARReceipt:
    receiptId: str
    agent: Union[AgentIdentity, Mapping[str, Any]]
    principal: Union[Principal, Mapping[str, Any]]
    action: Union[Action, Mapping[str, Any]]
    scope: Union[Scope, Mapping[str, Any]]
    inputHash: Union[HashObject, Mapping[str, Any]]
    outputHash: Union[HashObject, Mapping[str, Any]]
    timestamp: str
    cost: Union[Cost, Mapping[str, Any]]
    signature: Union[Signature, Mapping[str, Any]]
    metadata: Dict[str, Any]


@dataclass
class UnsignedReceipt:
    receiptId: str
    agent: Union[AgentIdentity, Mapping[str, Any]]
    principal: Union[Principal, Mapping[str, Any]]
    action: Union[Action, Mapping[str, Any]]
    scope: Union[Scope, Mapping[str, Any]]
    inputHash: Union[HashObject, Mapping[str, Any]]
    outputHash: Union[HashObject, Mapping[str, Any]]
    timestamp: str
    cost: Union[Cost, Mapping[str, Any]]
    signature: Union[UnsignedSignature, Mapping[str, Any]]
    metadata: Dict[str, Any]


@dataclass
class CreateReceiptOptions:
    agent: Union[AgentIdentity, Mapping[str, Any]]
    principal: Union[Principal, Mapping[str, Any]]
    action: Union[Action, Mapping[str, Any]]
    scope: Union[Scope, Mapping[str, Any]]
    inputHash: Union[HashObject, Mapping[str, Any]]
    outputHash: Union[HashObject, Mapping[str, Any]]
    cost: Union[Cost, Mapping[str, Any]]
    receiptId: Optional[str] = None
    timestamp: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    signature: Optional[Mapping[str, Any]] = None


@dataclass
class VerifyResult:
    ok: bool
    reason: Optional[str] = None


@dataclass
class KeyPair:
    publicKey: bytes
    secretKey: bytes


@dataclass
class WellKnownAARAgent:
    id: str
    publicKey: str
    name: Optional[str] = None
    version: Optional[str] = None


@dataclass
class WellKnownAARConfiguration:
    specVersion: str
    canonicalization: Literal["JCS-SORTED-UTF8-NOWS"]
    signatureAlgorithms: Tuple[Literal["Ed25519"], ...]
    hashAlgorithms: Tuple[Literal["sha256"], ...]
    receiptHeader: str
    agent: Union[WellKnownAARAgent, Mapping[str, Any]]


JSONScalar = Union[None, bool, int, float, str]
JSONValue = Union[JSONScalar, List["JSONValue"], Dict[str, "JSONValue"]]


def to_plain(value: Any) -> Any:
    """Recursively convert dataclasses to plain Python containers.

    Optional dataclass fields set to None are omitted so serialization behavior
    matches TS optional properties (undefined -> omitted by JSON.stringify).
    """
    if is_dataclass(value):
        out: Dict[str, Any] = {}
        extras: Dict[str, Any] = {}
        for dc_field in fields(value):
            item = getattr(value, dc_field.name)
            if dc_field.name == "extra":
                if isinstance(item, Mapping):
                    extras = {str(k): to_plain(v) for k, v in item.items()}
                continue
            if item is None:
                continue
            out[dc_field.name] = to_plain(item)
        out.update(extras)
        return out

    if isinstance(value, Mapping):
        return {str(k): to_plain(v) for k, v in value.items()}

    if isinstance(value, (list, tuple)):
        return [to_plain(v) for v in value]

    return value


def to_mapping(value: Union[Mapping[str, Any], Any], label: str) -> Dict[str, Any]:
    plain = to_plain(value)
    if not isinstance(plain, dict):
        raise TypeError("%s must be an object" % label)
    return plain


__all__ = [
    "AARReceipt",
    "Action",
    "ActionStatus",
    "AgentIdentity",
    "Cost",
    "CreateReceiptOptions",
    "HashAlgorithm",
    "HashObject",
    "JSONValue",
    "KeyPair",
    "Principal",
    "PrincipalType",
    "Scope",
    "Signature",
    "UnsignedReceipt",
    "UnsignedSignature",
    "VerifyResult",
    "WellKnownAARAgent",
    "WellKnownAARConfiguration",
    "to_plain",
    "to_mapping",
]



from __future__ import annotations

import re
from typing import Any, Dict, Mapping, Union

from nacl.signing import SigningKey

from .canonicalize import canonicalize_for_signing
from .encoding import decode_base64, decode_base64url, encode_base64url, utf8_encode
from .types import KeyPair, to_mapping, to_plain


def _sanitize_pem(pem: str) -> str:
    return re.sub(r"\s+", "", re.sub(r"-----END[^-]+-----", "", re.sub(r"-----BEGIN[^-]+-----", "", pem)))


def generate_key_pair() -> KeyPair:
    signing_key = SigningKey.generate()
    seed = signing_key.encode()
    public_key = signing_key.verify_key.encode()
    return KeyPair(publicKey=public_key, secretKey=seed + public_key)


def load_secret_key(input_key: Union[bytes, bytearray, str]) -> bytes:
    raw: bytes

    if isinstance(input_key, (bytes, bytearray)):
        raw = bytes(input_key)
    elif "-----BEGIN" in input_key:
        raw = decode_base64(_sanitize_pem(input_key))
    else:
        try:
            raw = decode_base64url(input_key)
        except ValueError:
            raw = decode_base64(input_key)

    if len(raw) == 64:
        return raw

    if len(raw) == 32:
        signing_key = SigningKey(raw)
        return raw + signing_key.verify_key.encode()

    raise ValueError("Secret key must be 32-byte seed or 64-byte Ed25519 secret key")


def public_key_from_secret(secret_key: Union[bytes, bytearray, str]) -> bytes:
    secret = load_secret_key(secret_key)
    seed = secret[:32]
    return SigningKey(seed).verify_key.encode()


def sign_receipt(unsigned: Union[Mapping[str, Any], Any], secret_key: Union[bytes, bytearray, str]) -> Dict[str, Any]:
    sk = load_secret_key(secret_key)
    unsigned_obj = to_mapping(unsigned, "unsigned")
    sig_obj = unsigned_obj.get("signature")

    if not isinstance(sig_obj, Mapping):
        raise ValueError("Invalid receipt.signature")

    fallback_public_key = encode_base64url(public_key_from_secret(sk))
    public_key = sig_obj["publicKey"] if "publicKey" in sig_obj and sig_obj["publicKey"] is not None else fallback_public_key

    with_defaults = {
        **unsigned_obj,
        "signature": {
            **to_plain(sig_obj),
            "alg": "Ed25519",
            "canonicalization": "JCS-SORTED-UTF8-NOWS",
            "publicKey": public_key,
        },
    }

    payload = utf8_encode(canonicalize_for_signing(with_defaults))
    signature = SigningKey(sk[:32]).sign(payload).signature

    return {
        **with_defaults,
        "signature": {
            **with_defaults["signature"],
            "sig": encode_base64url(signature),
        },
    }


generateKeyPair = generate_key_pair
loadSecretKey = load_secret_key
publicKeyFromSecret = public_key_from_secret
signReceipt = sign_receipt

__all__ = [
    "generate_key_pair",
    "generateKeyPair",
    "load_secret_key",
    "loadSecretKey",
    "public_key_from_secret",
    "publicKeyFromSecret",
    "sign_receipt",
    "signReceipt",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Union

from .encoding import encode_base64url
from .sign import public_key_from_secret


@dataclass
class DiscoveryOptions:
    agentId: str
    secretKey: Union[bytes, bytearray, str]
    agentName: Optional[str] = None
    agentVersion: Optional[str] = None
    receiptHeader: Optional[str] = None


def build_well_known_config(options: Union[DiscoveryOptions, Dict[str, Any]]) -> Dict[str, Any]:
    opts = options.__dict__ if isinstance(options, DiscoveryOptions) else dict(options)
    pk = public_key_from_secret(opts["secretKey"])

    return {
        "specVersion": "1.0",
        "canonicalization": "JCS-SORTED-UTF8-NOWS",
        "signatureAlgorithms": ["Ed25519"],
        "hashAlgorithms": ["sha256"],
        "receiptHeader": opts.get("receiptHeader") or "X-AAR-Receipt",
        "agent": {
            "id": opts["agentId"],
            "name": opts.get("agentName"),
            "version": opts.get("agentVersion"),
            "publicKey": encode_base64url(pk),
        },
    }


def well_known_handler(options: Union[DiscoveryOptions, Dict[str, Any]]) -> Callable[[Any, Any], None]:
    config = build_well_known_config(options)

    def _handler(_req: Any, res: Any) -> None:
        res.json(config)

    return _handler


buildWellKnownConfig = build_well_known_config
wellKnownHandler = well_known_handler

__all__ = [
    "DiscoveryOptions",
    "build_well_known_config",
    "buildWellKnownConfig",
    "well_known_handler",
    "wellKnownHandler",
]



from __future__ import annotations

import asyncio
import json
from typing import Any, Callable, Dict, List, Mapping, Optional, Union

from ..hash import hash_input, hash_output
from ..receipt import create_receipt, encode_receipt_header
from ..sign import load_secret_key, public_key_from_secret, sign_receipt

PrincipalResolver = Callable[[Any], Mapping[str, Any]]
ScopeResolver = Callable[[Any], List[str]]
CostResolver = Callable[[Any, Any], Optional[Mapping[str, Any]]]
PersistCallback = Callable[[Mapping[str, Any]], Any]


def _default_principal_resolver(request: Any) -> Dict[str, str]:
    wallet = request.headers.get("x-wallet")
    if wallet and wallet.strip():
        return {"id": wallet.strip(), "type": "user"}

    api_key = request.headers.get("x-api-key")
    if api_key and api_key.strip():
        return {"id": "apikey:%s..." % api_key.strip()[:8], "type": "service"}

    return {"id": "anonymous", "type": "other"}


def _default_scope_resolver(request: Any) -> List[str]:
    return ["%s:%s" % (request.method.lower(), request.path or "/")]


def _parse_body(body: bytes, content_type: str) -> Any:
    if not body:
        return None

    if "application/json" in content_type:
        try:
            return json.loads(body.decode("utf-8"))
        except Exception:
            return body.decode("utf-8", errors="replace")

    try:
        return body.decode("utf-8")
    except Exception:
        return body.hex()


def install_aar_middleware(
    app: Any,
    agent_id: str,
    secret_key: Union[bytes, bytearray, str],
    agent_name: Optional[str] = None,
    agent_version: Optional[str] = None,
    header_name: Optional[str] = None,
    principal_resolver: Optional[PrincipalResolver] = None,
    scope_resolver: Optional[ScopeResolver] = None,
    cost_resolver: Optional[CostResolver] = None,
    persist: Optional[PersistCallback] = None,
) -> None:
    try:
        from flask import g, request
    except Exception as exc:  # pragma: no cover - optional dependency path
        raise ImportError("flask is required to use install_aar_middleware") from exc

    sk = load_secret_key(secret_key)
    _pk = public_key_from_secret(sk)
    _ = _pk

    resolved_header = header_name or "X-AAR-Receipt"
    resolve_principal = principal_resolver or _default_principal_resolver
    resolve_scope = scope_resolver or _default_scope_resolver
    resolve_cost = cost_resolver or (lambda _req, _res: None)

    @app.before_request
    def _capture_body() -> None:
        g._aar_request_body = request.get_data(cache=True)

    @app.after_request
    def _attach_receipt(response: Any) -> Any:
        try:
            request_body = getattr(g, "_aar_request_body", b"")
            output_body = response.get_data() or b""

            input_data = {
                "query": request.args.to_dict(flat=True),
                "body": _parse_body(request_body, request.content_type or ""),
                "method": request.method,
                "path": request.path,
            }

            cost = resolve_cost(request, response) or {"amount": "0", "currency": "USD"}

            unsigned = create_receipt(
                {
                    "agent": {
                        "id": agent_id,
                        "name": agent_name,
                        "version": agent_version,
                    },
                    "principal": dict(resolve_principal(request)),
                    "action": {
                        "type": "api.call",
                        "target": request.path or "/",
                        "method": request.method.upper(),
                        "status": "success" if response.status_code < 400 else "failure",
                    },
                    "scope": {"permissions": list(resolve_scope(request))},
                    "inputHash": hash_input(input_data),
                    "outputHash": hash_output(output_body),
                    "cost": dict(cost),
                }
            )

            receipt = sign_receipt(unsigned, sk)
            if resolved_header not in response.headers:
                response.headers[resolved_header] = encode_receipt_header(receipt)

            if persist is not None:
                result = persist(receipt)
                if hasattr(result, "__await__"):
                    try:
                        loop = asyncio.get_running_loop()
                        loop.create_task(result)
                    except RuntimeError:
                        pass
        except Exception:
            pass

        return response


aar_middleware = install_aar_middleware

__all__ = ["install_aar_middleware", "aar_middleware"]



from __future__ import annotations

import pytest
from botindex_aar.sign import generate_key_pair, load_secret_key, sign_receipt
from botindex_aar.verify import verify_receipt
from botindex_aar.receipt import create_receipt
from botindex_aar.hash import hash_input, hash_output


def test_generate_key_pair() -> None:
    keys = generate_key_pair()
    assert len(bytes(keys.publicKey)) == 32
    assert len(keys.secretKey) == 64  # PyNaCl expanded secret key


def test_sign_and_verify_roundtrip() -> None:
    keys = generate_key_pair()
    receipt = create_receipt({
        "receiptId": "test-001",
        "timestamp": "2026-03-06T12:00:00.000Z",
        "agent": {"id": "agent:test", "name": "TestAgent", "version": "1.0.0"},
        "principal": {"id": "user:test", "type": "user"},
        "action": {"type": "test.action", "target": "unit-test", "method": "GET", "status": "success"},
        "scope": {"permissions": ["test.read"], "constraints": {}},
        "inputHash": hash_input({"query": "test"}),
        "outputHash": hash_output("test-result"),
        "cost": {"amount": "0.01", "currency": "USDC"},
        "signature": {"kid": "agent:test#key-1"},
        "metadata": {},
    })
    signed = sign_receipt(receipt, keys.secretKey)
    assert "sig" in signed["signature"]
    assert signed["signature"]["alg"] == "Ed25519"
    assert signed["signature"]["canonicalization"] == "JCS-SORTED-UTF8-NOWS"

    result = verify_receipt(signed, keys.publicKey)
    assert result.ok is True


def test_verify_fails_with_wrong_key() -> None:
    keys1 = generate_key_pair()
    keys2 = generate_key_pair()
    receipt = create_receipt({
        "receiptId": "test-002",
        "timestamp": "2026-03-06T12:00:00.000Z",
        "agent": {"id": "agent:test", "name": "TestAgent", "version": "1.0.0"},
        "principal": {"id": "user:test", "type": "user"},
        "action": {"type": "test.action", "target": "unit-test", "method": "GET", "status": "success"},
        "scope": {"permissions": ["test.read"], "constraints": {}},
        "inputHash": hash_input({"query": "test"}),
        "outputHash": hash_output("test-result"),
        "cost": {"amount": "0.01", "currency": "USDC"},
        "signature": {"kid": "agent:test#key-1"},
        "metadata": {},
    })
    signed = sign_receipt(receipt, keys1.secretKey)
    result = verify_receipt(signed, keys2.publicKey)
    assert result.ok is False


def test_verify_fails_with_tampered_receipt() -> None:
    keys = generate_key_pair()
    receipt = create_receipt({
        "receiptId": "test-003",
        "timestamp": "2026-03-06T12:00:00.000Z",
        "agent": {"id": "agent:test", "name": "TestAgent", "version": "1.0.0"},
        "principal": {"id": "user:test", "type": "user"},
        "action": {"type": "test.action", "target": "unit-test", "method": "GET", "status": "success"},
        "scope": {"permissions": ["test.read"], "constraints": {}},
        "inputHash": hash_input({"query": "test"}),
        "outputHash": hash_output("test-result"),
        "cost": {"amount": "0.01", "currency": "USDC"},
        "signature": {"kid": "agent:test#key-1"},
        "metadata": {},
    })
    signed = sign_receipt(receipt, keys.secretKey)
    # Tamper with the receipt
    signed["action"]["target"] = "tampered"
    result = verify_receipt(signed, keys.publicKey)
    assert result.ok is False


def test_load_secret_key_from_seed() -> None:
    seed = bytes(range(32))
    sk = load_secret_key(seed)
    assert sk is not None  # Expanded Ed25519 secret key



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, List, Mapping, Optional, Union

from ..hash import hash_input, hash_output
from ..receipt import create_receipt, encode_receipt_header
from ..sign import load_secret_key, public_key_from_secret, sign_receipt

try:
    from starlette.middleware.base import BaseHTTPMiddleware
    from starlette.requests import Request
    from starlette.responses import Response

    _STARLETTE_AVAILABLE = True
except Exception:  # pragma: no cover - optional dependency path
    BaseHTTPMiddleware = object  # type: ignore[assignment]
    Request = Any  # type: ignore[misc,assignment]
    Response = Any  # type: ignore[misc,assignment]
    _STARLETTE_AVAILABLE = False


PrincipalResolver = Callable[[Request], Mapping[str, Any]]
ScopeResolver = Callable[[Request], List[str]]
CostResolver = Callable[[Request, Response], Optional[Mapping[str, Any]]]
PersistCallback = Callable[[Mapping[str, Any]], Union[None, Awaitable[None]]]


@dataclass
class AARMiddlewareOptions:
    agent_id: str
    secret_key: Union[bytes, bytearray, str]
    agent_name: Optional[str] = None
    agent_version: Optional[str] = None
    header_name: Optional[str] = None
    principal_resolver: Optional[PrincipalResolver] = None
    scope_resolver: Optional[ScopeResolver] = None
    cost_resolver: Optional[CostResolver] = None
    persist: Optional[PersistCallback] = None


def _default_principal_resolver(request: Request) -> Dict[str, str]:
    wallet = request.headers.get("x-wallet")
    if wallet and wallet.strip():
        return {"id": wallet.strip(), "type": "user"}

    api_key = request.headers.get("x-api-key")
    if api_key and api_key.strip():
        return {"id": "apikey:%s..." % api_key.strip()[:8], "type": "service"}

    return {"id": "anonymous", "type": "other"}


def _default_scope_resolver(request: Request) -> List[str]:
    method = request.method.lower()
    path = request.url.path or "/"
    return ["%s:%s" % (method, path)]


def _parse_body(body: bytes, content_type: str) -> Any:
    if not body:
        return None

    if "application/json" in content_type:
        try:
            return json.loads(body.decode("utf-8"))
        except Exception:
            return body.decode("utf-8", errors="replace")

    try:
        return body.decode("utf-8")
    except Exception:
        return body.hex()


class AARMiddleware(BaseHTTPMiddleware):
    """FastAPI/Starlette middleware that emits signed AAR receipts on responses."""

    def __init__(
        self,
        app: Any,
        agent_id: str,
        secret_key: Union[bytes, bytearray, str],
        agent_name: Optional[str] = None,
        agent_version: Optional[str] = None,
        header_name: Optional[str] = None,
        principal_resolver: Optional[PrincipalResolver] = None,
        scope_resolver: Optional[ScopeResolver] = None,
        cost_resolver: Optional[CostResolver] = None,
        persist: Optional[PersistCallback] = None,
    ) -> None:
        if not _STARLETTE_AVAILABLE:  # pragma: no cover - optional dependency path
            raise ImportError("starlette/fastapi is required to use AARMiddleware")

        super().__init__(app)
        self._secret_key = load_secret_key(secret_key)
        self._public_key = public_key_from_secret(self._secret_key)
        self._agent_id = agent_id
        self._agent_name = agent_name
        self._agent_version = agent_version
        self._header_name = header_name or "X-AAR-Receipt"
        self._principal_resolver = principal_resolver or _default_principal_resolver
        self._scope_resolver = scope_resolver or _default_scope_resolver
        self._cost_resolver = cost_resolver or (lambda _req, _res: None)
        self._persist = persist

    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        body = await request.body()

        sent = False

        async def receive() -> Dict[str, Any]:
            nonlocal sent
            if sent:
                return {"type": "http.request", "body": b"", "more_body": False}
            sent = True
            return {"type": "http.request", "body": body, "more_body": False}

        req = Request(request.scope, receive)
        response = await call_next(req)

        output_chunks: List[bytes] = []
        async for chunk in response.body_iterator:
            output_chunks.append(chunk if isinstance(chunk, bytes) else bytes(chunk))
        output_body = b"".join(output_chunks)

        wrapped = Response(
            content=output_body,
            status_code=response.status_code,
            headers=dict(response.headers),
            media_type=response.media_type,
            background=response.background,
        )

        try:
            input_data = {
                "query": dict(req.query_params),
                "body": _parse_body(body, req.headers.get("content-type", "")),
                "method": req.method,
                "path": req.url.path,
            }

            cost = self._cost_resolver(req, wrapped) or {"amount": "0", "currency": "USD"}
            action_path = req.url.path or "/"

            unsigned = create_receipt(
                {
                    "agent": {
                        "id": self._agent_id,
                        "name": self._agent_name,
                        "version": self._agent_version,
                    },
                    "principal": dict(self._principal_resolver(req)),
                    "action": {
                        "type": "api.call",
                        "target": action_path,
                        "method": req.method.upper(),
                        "status": "success" if wrapped.status_code < 400 else "failure",
                    },
                    "scope": {"permissions": list(self._scope_resolver(req))},
                    "inputHash": hash_input(input_data),
                    "outputHash": hash_output(output_body),
                    "cost": dict(cost),
                }
            )

            receipt = sign_receipt(unsigned, self._secret_key)
            if self._header_name not in wrapped.headers:
                wrapped.headers[self._header_name] = encode_receipt_header(receipt)

            if self._persist is not None:
                maybe_awaitable = self._persist(receipt)
                if hasattr(maybe_awaitable, "__await__"):
                    await maybe_awaitable  # type: ignore[misc]
        except Exception:
            pass

        return wrapped


__all__ = ["AARMiddleware", "AARMiddlewareOptions"]



from __future__ import annotations

import json
from typing import Any, Dict, Mapping

from .types import to_plain


def _is_object(value: Any) -> bool:
    return isinstance(value, Mapping)


def _sort_keys_deep(value: Any) -> Any:
    if isinstance(value, list):
        return [_sort_keys_deep(item) for item in value]

    if not _is_object(value):
        return value

    sorted_obj: Dict[str, Any] = {}
    for key in sorted(value.keys()):
        sorted_obj[str(key)] = _sort_keys_deep(value[key])
    return sorted_obj


def canonicalize(value: Any) -> str:
    normalized = to_plain(value)
    return json.dumps(_sort_keys_deep(normalized), ensure_ascii=False, separators=(",", ":"))


def _to_signable_receipt(receipt: Any) -> Dict[str, Any]:
    clone_json = json.dumps(to_plain(receipt), ensure_ascii=False, separators=(",", ":"))
    clone_obj = json.loads(clone_json)

    signature = clone_obj.get("signature") if isinstance(clone_obj, dict) else None
    if not isinstance(signature, dict):
        raise ValueError("Invalid receipt.signature")

    signature.pop("sig", None)
    clone_obj["signature"] = signature
    return clone_obj


def canonicalize_for_signing(receipt: Any) -> str:
    return canonicalize(_to_signable_receipt(receipt))


canonicalizeForSigning = canonicalize_for_signing

__all__ = ["canonicalize", "canonicalize_for_signing", "canonicalizeForSigning"]



from __future__ import annotations

import hashlib

from botindex_aar.hash import hash_input, hash_output, sha256


def test_sha256_matches_hashlib() -> None:
    assert sha256(b"abc") == hashlib.sha256(b"abc").digest()


def test_hash_output_string_digest() -> None:
    result = hash_output("abc")
    assert result.alg == "sha256"
    assert result.digest == "ungWv48Bz-pBQUDeXa4iI7ADYaOWF3qctBD_YfIAFa0"


def test_hash_input_canonicalized_digest() -> None:
    result = hash_input({"b": 1, "a": 2, "nested": {"z": 0, "y": [3, {"b": 2, "a": 1}]}})
    assert result.alg == "sha256"
    assert result.digest == "WMp5KcCXKZ68aTP4RRbMHMcHjfmNrdy6IgQHg8-ghgU"
